/*
 * Copyright (c) 2021 - 2022, Shanghai Yunsilicon Technology Co., Ltd.
 * All rights reserved.
 */

#ifndef _XSCDV_H_
#define _XSCDV_H_

#include <stdio.h>
#include <linux/types.h> /* For the __be64 type */
#include <sys/types.h>
#include <endian.h>
#if defined(__SSE3__)
#include <limits.h>
#include <emmintrin.h>
#include <tmmintrin.h>
#endif /* defined(__SSE3__) */

#include <infiniband/verbs.h>
#include <infiniband/tm_types.h>
#include <infiniband/xsc_api.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Always inline the functions */
#ifdef __GNUC__
#define XSCDV_ALWAYS_INLINE inline __attribute__((always_inline))
#else
#define XSCDV_ALWAYS_INLINE inline
#endif

enum {
	XSC_RCV_DBR	= 0,
	XSC_SND_DBR	= 1,
};

enum xscdv_context_comp_mask {
	XSCDV_CONTEXT_MASK_CQE_COMPRESION	= 1 << 0,
	XSCDV_CONTEXT_MASK_SWP			= 1 << 1,
	XSCDV_CONTEXT_MASK_STRIDING_RQ		= 1 << 2,
	XSCDV_CONTEXT_MASK_TUNNEL_OFFLOADS	= 1 << 3,
	XSCDV_CONTEXT_MASK_DYN_BFREGS		= 1 << 4,
	XSCDV_CONTEXT_MASK_CLOCK_INFO_UPDATE	= 1 << 5,
	XSCDV_CONTEXT_MASK_FLOW_ACTION_FLAGS	= 1 << 6,
};

struct xscdv_cqe_comp_caps {
	uint32_t max_num;
	uint32_t supported_format; /* enum xscdv_cqe_comp_res_format */
};

struct xscdv_sw_parsing_caps {
	uint32_t sw_parsing_offloads; /* Use enum xscdv_sw_parsing_offloads */
	uint32_t supported_qpts;
};

struct xscdv_striding_rq_caps {
	uint32_t min_single_stride_log_num_of_bytes;
	uint32_t max_single_stride_log_num_of_bytes;
	uint32_t min_single_wqe_log_num_of_strides;
	uint32_t max_single_wqe_log_num_of_strides;
	uint32_t supported_qpts;
};

/*
 * Direct verbs device-specific attributes
 */
struct xscdv_context {
	uint8_t		version;
	uint64_t	flags;
	uint64_t	comp_mask;
	struct xscdv_cqe_comp_caps	cqe_comp_caps;
	struct xscdv_sw_parsing_caps sw_parsing_caps;
	struct xscdv_striding_rq_caps striding_rq_caps;
	uint32_t	tunnel_offloads_caps;
	uint64_t	max_clock_info_update_nsec;
	uint32_t        flow_action_flags;
};

enum xscsdv_context_flags {
	/*
	 * This flag indicates if CQE version 0 or 1 is needed.
	 */
	XSCDV_CONTEXT_FLAGS_CQE_V1	= (1 << 0),
	XSCDV_CONTEXT_FLAGS_OBSOLETE	= (1 << 1), /* Obsoleted, don't use */
	XSCDV_CONTEXT_FLAGS_MPW_ALLOWED = (1 << 2),
	XSCDV_CONTEXT_FLAGS_ENHANCED_MPW = (1 << 3),
	XSCDV_CONTEXT_FLAGS_CQE_128B_COMP = (1 << 4), /* Support CQE 128B compression */
	XSCDV_CONTEXT_FLAGS_CQE_128B_PAD = (1 << 5), /* Support CQE 128B padding */
};

enum xscdv_cq_init_attr_mask {
	XSCDV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE	= 1 << 0,
	XSCDV_CQ_INIT_ATTR_MASK_FLAGS		= 1 << 1,
	XSCDV_CQ_INIT_ATTR_MASK_CQE_SIZE = 1 << 2,
};

struct xscdv_cq_init_attr {
	uint64_t comp_mask; /* Use enum xscdv_cq_init_attr_mask */
	uint8_t cqe_comp_res_format; /* Use enum xscdv_cqe_comp_res_format */
	uint32_t flags;
	uint16_t cqe_size; /* when XSCDV_CQ_INIT_ATTR_MASK_CQE_SIZE set */
};

struct ibv_cq_ex *xscdv_create_cq(struct ibv_context *context,
				   struct ibv_cq_init_attr_ex *cq_attr,
				   struct xscdv_cq_init_attr *xcq_attr);

enum xscdv_qp_create_flags {
	XSCDV_QP_CREATE_TUNNEL_OFFLOADS = 1 << 0,
	XSCDV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC = 1 << 1,
	XSCDV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC = 1 << 2,
	XSCDV_QP_CREATE_DISABLE_SCATTER_TO_CQE = 1 << 3,
	XSCDV_QP_CREATE_ALLOW_SCATTER_TO_CQE = 1 << 4,
};

enum xscdv_qp_init_attr_mask {
	XSCDV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS	= 1 << 0,
	XSCDV_QP_INIT_ATTR_MASK_DC			= 1 << 1,
};

enum xscdv_dc_type {
	XSCDV_DCTYPE_DCT     = 1,
	XSCDV_DCTYPE_DCI,
};

struct xscdv_dc_init_attr {
	enum xscdv_dc_type	dc_type;
	uint64_t dct_access_key;
};

struct xscdv_qp_init_attr {
	uint64_t comp_mask;	/* Use enum xscdv_qp_init_attr_mask */
	uint32_t create_flags;	/* Use enum xsc_qp_create_flags */
	struct xscdv_dc_init_attr  dc_init_attr;
};

struct ibv_qp *xscdv_create_qp(struct ibv_context *context,
				struct ibv_qp_init_attr_ex *qp_attr,
				struct xscdv_qp_init_attr *xqp_attr);

enum xscdv_flow_action_esp_mask {
	XSCDV_FLOW_ACTION_ESP_MASK_FLAGS	= 1 << 0,
};

struct xscdv_flow_action_esp {
	uint64_t comp_mask;  /* Use enum xscdv_flow_action_esp_mask */
	uint32_t action_flags; /* Use enum xscdv_flow_action_flags */
};

struct xscdv_flow_match_parameters {
	size_t match_sz;
	uint64_t match_buf[]; /* Device spec format */
};

struct xscdv_flow_matcher_attr {
	enum ibv_flow_attr_type type;
	uint32_t flags; /* From enum ibv_flow_flags */
	uint16_t priority;
	uint8_t match_criteria_enable; /* Device spec format */
	struct xscdv_flow_match_parameters *match_mask;
	uint64_t comp_mask;
};

struct xscdv_flow_matcher;

struct xscdv_flow_matcher *
xscdv_create_flow_matcher(struct ibv_context *context,
			   struct xscdv_flow_matcher_attr *matcher_attr);

int xscdv_destroy_flow_matcher(struct xscdv_flow_matcher *matcher);

enum xscdv_flow_action_type {
	XSCDV_FLOW_ACTION_DEST_IBV_QP,
	XSCDV_FLOW_ACTION_DROP,
	XSCDV_FLOW_ACTION_IBV_COUNTER,
	XSCDV_FLOW_ACTION_IBV_FLOW_ACTION,
	XSCDV_FLOW_ACTION_TAG,
	XSCDV_FLOW_ACTION_DEST_DEVX,
};

struct xscdv_flow_action_attr {
	enum xscdv_flow_action_type type;
	union {
		struct ibv_qp *qp;
		struct ibv_counters *counter;
		struct ibv_flow_action *action;
		uint32_t tag_value;
		struct xscdv_devx_obj *obj;
	};
};

struct ibv_flow *
xscdv_create_flow(struct xscdv_flow_matcher *matcher,
		   struct xscdv_flow_match_parameters *match_value,
		   size_t num_actions,
		   struct xscdv_flow_action_attr actions_attr[]);

struct ibv_flow_action *xscdv_create_flow_action_esp(struct ibv_context *ctx,
						      struct ibv_flow_action_esp_attr *esp,
						      struct xscdv_flow_action_esp *xattr);

/*
 * xscdv_create_flow_action_modify_header - Create a flow action which mutates
 * a packet. The flow action can be attached to steering rules via
 * ibv_create_flow().
 *
 * @ctx: RDMA device context to create the action on.
 * @actions_sz: The size of *actions* buffer in bytes.
 * @actions: A buffer which contains modify actions provided in device spec
 *	     format.
 * @ft_type: Defines the flow table type to which the modify
 *	     header action will be attached.
 *
 * Return a valid ibv_flow_action if successful, NULL otherwise.
 */
struct ibv_flow_action *
xscdv_create_flow_action_modify_header(struct ibv_context *ctx,
					size_t actions_sz,
					uint64_t actions[],
					enum xscdv_flow_table_type ft_type);

/*
 * xscdv_create_flow_action_packet_reformat - Create flow action which can
 * encap/decap packets.
 */
struct ibv_flow_action *
xscdv_create_flow_action_packet_reformat(struct ibv_context *ctx,
					  size_t data_sz,
					  void *data,
					  enum xscdv_flow_action_packet_reformat_type reformat_type,
					  enum xscdv_flow_table_type ft_type);
/*
 * Most device capabilities are exported by ibv_query_device(...),
 * but there is HW device-specific information which is important
 * for data-path, but isn't provided.
 *
 * Return 0 on success.
 */
int xscdv_query_device(struct ibv_context *ctx_in,
			struct xscdv_context *attrs_out);

enum xscdv_qp_comp_mask {
	XSCDV_QP_MASK_UAR_MMAP_OFFSET		= 1 << 0,
	XSCDV_QP_MASK_RAW_QP_HANDLES		= 1 << 1,
};

struct xscdv_qp {
	__le32			*dbrec;
	struct {
		void		*buf;
		uint32_t	wqe_cnt;
		uint32_t	stride;
		__le32		*db;
	} sq;
	struct {
		void		*buf;
		uint32_t	wqe_cnt;
		uint32_t	stride;
		__le32		*db;
	} rq;
	uint64_t		comp_mask;
	uint32_t		tirn;
	uint32_t		tisn;
	uint32_t		rqn;
	uint32_t		sqn;
};

struct xscdv_cq {
	void			*buf;
	__le32			*dbrec;
	__le32			*db;
	uint32_t		cqe_cnt;
	uint32_t		cqe_size;
	uint32_t		cqn;
	uint64_t		comp_mask;
};

struct xscdv_rwq {
	void		*buf;
	__le32		*dbrec;
	uint32_t	wqe_cnt;
	uint32_t	stride;
	uint64_t	comp_mask;
	__le32		*db;
};

struct xscdv_dm {
	void		*buf;
	uint64_t	length;
	uint64_t	comp_mask;
};

struct xsc_wqe_av;

struct xscdv_ah {
	struct xsc_wqe_av      *av;
	uint64_t		comp_mask;
};

struct xscdv_pd {
	uint32_t		pdn;
	uint64_t		comp_mask;
};

struct xscdv_obj {
	struct {
		struct ibv_qp		*in;
		struct xscdv_qp		*out;
	} qp;
	struct {
		struct ibv_cq		*in;
		struct xscdv_cq	*out;
	} cq;
	struct {
		struct ibv_wq		*in;
		struct xscdv_rwq	*out;
	} rwq;
	struct {
		struct ibv_dm		*in;
		struct xscdv_dm	*out;
	} dm;
	struct {
		struct ibv_ah		*in;
		struct xscdv_ah	*out;
	} ah;
	struct {
		struct ibv_pd		*in;
		struct xscdv_pd	*out;
	} pd;
};

enum xscdv_obj_type {
	XSCDV_OBJ_QP	= 1 << 0,
	XSCDV_OBJ_CQ	= 1 << 1,
	XSCDV_OBJ_SRQ	= 1 << 2,
	XSCDV_OBJ_RWQ	= 1 << 3,
	XSCDV_OBJ_DM	= 1 << 4,
	XSCDV_OBJ_AH	= 1 << 5,
	XSCDV_OBJ_PD	= 1 << 6,
};

enum xscdv_wq_init_attr_mask {
	XSCDV_WQ_INIT_ATTR_MASK_STRIDING_RQ	= 1 << 0,
};

struct xscdv_striding_rq_init_attr {
	uint32_t	single_stride_log_num_of_bytes;
	uint32_t	single_wqe_log_num_of_strides;
	uint8_t		two_byte_shift_en;
};

struct xscdv_wq_init_attr {
	uint64_t				comp_mask; /* Use enum xscdv_wq_init_attr_mask */
	struct xscdv_striding_rq_init_attr	striding_rq_attrs;
};

/*
 * This function creates a work queue object with extra properties
 * defined by xscdv_wq_init_attr struct.
 *
 * For each bit in the comp_mask, a field in xscdv_wq_init_attr
 * should follow.
 *
 * XSCDV_WQ_INIT_ATTR_MASK_STRIDING_RQ: Create a work queue with
 * striding RQ capabilities.
 * - single_stride_log_num_of_bytes represents the size of each stride in the
 *   WQE and its value should be between min_single_stride_log_num_of_bytes
 *   and max_single_stride_log_num_of_bytes that are reported in
 *   xscdv_query_device.
 * - single_wqe_log_num_of_strides represents the number of strides in each WQE.
 *   Its value should be between min_single_wqe_log_num_of_strides and
 *   max_single_wqe_log_num_of_strides that are reported in xscdv_query_device.
 * - two_byte_shift_en: When enabled, hardware pads 2 bytes of zeroes
 *   before writing the message to memory (e.g. for IP alignment)
 */
struct ibv_wq *xscdv_create_wq(struct ibv_context *context,
				struct ibv_wq_init_attr *wq_init_attr,
				struct xscdv_wq_init_attr *xwq_attr);
/*
 * This function will initialize xscdv_xxx structs based on supplied type.
 * The information for initialization is taken from ibv_xx structs supplied
 * as part of input.
 *
 * Request information of CQ marks its owned by DV for all consumer index
 * related actions.
 *
 * The initialization type can be combination of several types together.
 *
 * Return: 0 in case of success.
 */
int xscdv_init_obj(struct xscdv_obj *obj, uint64_t obj_type);

enum {
	XSC_OPCODE_NOP			= 0x00,
	XSC_OPCODE_SEND_INVAL		= 0x01,
	XSC_OPCODE_RDMA_WRITE		= 0x08,
	XSC_OPCODE_RDMA_WRITE_IMM	= 0x09,
	XSC_OPCODE_SEND		= 0x0a,
	XSC_OPCODE_SEND_IMM		= 0x0b,
	XSC_OPCODE_TSO			= 0x0e,
	XSC_OPCODE_RDMA_READ		= 0x10,
	XSC_OPCODE_ATOMIC_CS		= 0x11,
	XSC_OPCODE_ATOMIC_FA		= 0x12,
	XSC_OPCODE_ATOMIC_MASKED_CS	= 0x14,
	XSC_OPCODE_ATOMIC_MASKED_FA	= 0x15,
	XSC_OPCODE_FMR			= 0x19,
	XSC_OPCODE_LOCAL_INVAL		= 0x1b,
	XSC_OPCODE_CONFIG_CMD		= 0x1f,
	XSC_OPCODE_UMR			= 0x25,
	XSC_OPCODE_TAG_MATCHING	= 0x28
};

enum {
	XSC_CQE_L2_OK = 1 << 0,
	XSC_CQE_L3_OK = 1 << 1,
	XSC_CQE_L4_OK = 1 << 2,
};

enum {
	XSC_CQE_L3_HDR_TYPE_NONE = 0x0,
	XSC_CQE_L3_HDR_TYPE_IPV6 = 0x1,
	XSC_CQE_L3_HDR_TYPE_IPV4 = 0x2,
};

enum {
	XSC_CQE_OWNER_MASK	= 1,
	XSC_CQE_REQ		= 0,
	XSC_CQE_RESP_WR_IMM	= 1,
	XSC_CQE_RESP_SEND	= 2,
	XSC_CQE_RESP_SEND_IMM	= 3,
	XSC_CQE_RESP_SEND_INV	= 4,
	XSC_CQE_RESIZE_CQ	= 5,
	XSC_CQE_NO_PACKET	= 6,
	XSC_CQE_REQ_ERR	= 13,
	XSC_CQE_RESP_ERR	= 14,
	XSC_CQE_INVALID	= 15,
};

struct xsc_err_cqe {
	uint8_t		rsvd0[32];
	uint32_t	srqn;
	uint8_t		rsvd1[18];
	uint8_t		vendor_err_synd;
	uint8_t		syndrome;
	uint32_t	s_wqe_opcode_qpn;
	uint16_t	wqe_counter;
	uint8_t		signature;
	uint8_t		op_own;
};

struct xsc_tm_cqe {
	__be32		success;
	__be16		hw_phase_cnt;
	uint8_t		rsvd0[12];
};

struct xsc_cqe64 {
	union {
		struct {
			uint8_t		rsvd0[2];
			__be16		wqe_id;
			uint8_t		rsvd4[13];
			uint8_t		ml_path;
			uint8_t		rsvd20[4];
			__be16		slid;
			__be32		flags_rqpn;
			uint8_t		hds_ip_ext;
			uint8_t		l4_hdr_type_etc;
			__be16		vlan_info;
		};
		struct xsc_tm_cqe tm_cqe;
		/* TMH is scattered to CQE upon match */
		struct ibv_tmh tmh;
	};
	__be32		srqn_uidx;
	__be32		imm_inval_pkey;
	uint8_t		app;
	uint8_t		app_op;
	__be16		app_info;
	__be32		byte_cnt;
	__be64		timestamp;
	__be32		sop_drop_qpn;
	__be16		wqe_counter;
	uint8_t		signature;
	uint8_t		op_own;
};

enum xscdv_cqe_comp_res_format {
	XSCDV_CQE_RES_FORMAT_HASH		= 1 << 0,
	XSCDV_CQE_RES_FORMAT_CSUM		= 1 << 1,
	XSCDV_CQE_RES_FORMAT_CSUM_STRIDX       = 1 << 2,
};

enum xscdv_sw_parsing_offloads {
	XSCDV_SW_PARSING		= 1 << 0,
	XSCDV_SW_PARSING_CSUM		= 1 << 1,
	XSCDV_SW_PARSING_LSO		= 1 << 2,
};

static XSCDV_ALWAYS_INLINE
uint8_t xscdv_get_cqe_owner(struct xsc_cqe64 *cqe)
{
	return cqe->op_own & 0x1;
}

static XSCDV_ALWAYS_INLINE
void xscdv_set_cqe_owner(struct xsc_cqe64 *cqe, uint8_t val)
{
	cqe->op_own = (val & 0x1) | (cqe->op_own & ~0x1);
}

/* Solicited event */
static XSCDV_ALWAYS_INLINE
uint8_t xscdv_get_cqe_se(struct xsc_cqe64 *cqe)
{
	return (cqe->op_own >> 1) & 0x1;
}

static XSCDV_ALWAYS_INLINE
uint8_t xscdv_get_cqe_format(struct xsc_cqe64 *cqe)
{
	return (cqe->op_own >> 2) & 0x3;
}

static XSCDV_ALWAYS_INLINE
uint8_t xscdv_get_cqe_opcode(struct xsc_cqe64 *cqe)
{
	return cqe->op_own >> 4;
}

/*
 * WQE related part
 */
enum {
	XSC_INVALID_LKEY	= 0x100,
};

enum {
	XSC_SEND_WQE_BB	= 64,
	XSC_SEND_WQE_SHIFT	= 6,
};

struct xsc_wqe_srq_next_seg {
	uint8_t			rsvd0[2];
	__be16			next_wqe_index;
	uint8_t			signature;
	uint8_t			rsvd1[11];
};

struct xsc_wqe_ctrl_seg {
	__be32		opmod_idx_opcode;
	__be32		qpn_ds;
	uint8_t		signature;
	uint8_t		rsvd[2];
	uint8_t		fm_ce_se;
	__be32		imm;
};

struct xsc_wqe_av {
	union {
		struct {
			__be32		qkey;
			__be32		reserved;
		} qkey;
		__be64		dc_key;
	} key;
	__be32		dqp_dct;
	uint8_t		stat_rate_sl;
	uint8_t		fl_mlid;
	__be16		rlid;
	uint8_t		reserved0[4];
	uint8_t		rmac[6];
	uint8_t		tclass;
	uint8_t		hop_limit;
	__be32		grh_gid_fl;
	uint8_t		rgid[16];
};

struct xsc_wqe_datagram_seg {
	struct xsc_wqe_av	av;
};

struct xsc_wqe_raddr_seg {
	__be64		raddr;
	__be32		rkey;
	__be32		reserved;
};

struct xsc_wqe_atomic_seg {
	__be64		swap_add;
	__be64		compare;
};

struct xsc_wqe_inl_data_seg {
	uint32_t	byte_count;
};

struct xsc_wqe_eth_seg {
	__be32		rsvd0;
	uint8_t		cs_flags;
	uint8_t		rsvd1;
	__be16		mss;
	__be32		rsvd2;
	__be16		inline_hdr_sz;
	uint8_t		inline_hdr_start[2];
	uint8_t		inline_hdr[16];
};

/*
 * Control segment - contains some control information for the current WQE.
 *
 * Output:
 *	seg	  - control segment to be filled
 * Input:
 *	pi	  - WQEBB number of the first block of this WQE.
 *		    This number should wrap at 0xffff, regardless of
 *		    size of the WQ.
 *	opcode	  - Opcode of this WQE. Encodes the type of operation
 *		    to be executed on the QP.
 *	opmod	  - Opcode modifier.
 *	qp_num	  - QP/SQ number this WQE is posted to.
 *	fm_ce_se  - FM (fence mode), CE (completion and event mode)
 *		    and SE (solicited event).
 *	ds	  - WQE size in octowords (16-byte units). DS accounts for all
 *		    the segments in the WQE as summarized in WQE construction.
 *	signature - WQE signature.
 *	imm	  - Immediate data/Invalidation key/UMR mkey.
 */
static XSCDV_ALWAYS_INLINE
void xscdv_set_ctrl_seg(struct xsc_wqe_ctrl_seg *seg, uint16_t pi,
			 uint8_t opcode, uint8_t opmod, uint32_t qp_num,
			 uint8_t fm_ce_se, uint8_t ds,
			 uint8_t signature, uint32_t imm)
{
	seg->opmod_idx_opcode	= htobe32(((uint32_t)opmod << 24) | ((uint32_t)pi << 8) | opcode);
	seg->qpn_ds		= htobe32((qp_num << 8) | ds);
	seg->fm_ce_se		= fm_ce_se;
	seg->signature		= signature;
	/*
	 * The caller should prepare "imm" in advance based on WR opcode.
	 * For IBV_WR_SEND_WITH_IMM and IBV_WR_RDMA_WRITE_WITH_IMM,
	 * the "imm" should be assigned as is.
	 * For the IBV_WR_SEND_WITH_INV, it should be htobe32(imm).
	 */
	seg->imm		= imm;
}

/* x86 optimized version of xscdv_set_ctrl_seg()
 *
 * This is useful when doing calculations on large data sets
 * for parallel calculations.
 *
 * It doesn't suit for serialized algorithms.
 */
#if defined(__SSE3__)
static XSCDV_ALWAYS_INLINE
void xscdv_x86_set_ctrl_seg(struct xsc_wqe_ctrl_seg *seg, uint16_t pi,
			     uint8_t opcode, uint8_t opmod, uint32_t qp_num,
			     uint8_t fm_ce_se, uint8_t ds,
			     uint8_t signature, uint32_t imm)
{
	__m128i val  = _mm_set_epi32(imm, qp_num, (ds << 16) | pi,
				     (signature << 24) | (opcode << 16) | (opmod << 8) | fm_ce_se);
	__m128i mask = _mm_set_epi8(15, 14, 13, 12,	/* immediate */
				     0,			/* signal/fence_mode */
#if CHAR_MIN
				     -128, -128,        /* reserved */
#else
				     0x80, 0x80,        /* reserved */
#endif
				     3,			/* signature */
				     6,			/* data size */
				     8, 9, 10,		/* QP num */
				     2,			/* opcode */
				     4, 5,		/* sw_pi in BE */
				     1			/* opmod */
				     );
	*(__m128i *) seg = _mm_shuffle_epi8(val, mask);
}
#endif /* defined(__SSE3__) */

/*
 * Datagram Segment - contains address information required in order
 * to form a datagram message.
 *
 * Output:
 *	seg		- datagram segment to be filled.
 * Input:
 *	key		- Q_key/access key.
 *	dqp_dct		- Destination QP number for UD and DCT for DC.
 *	ext		- Address vector extension.
 *	stat_rate_sl	- Maximum static rate control, SL/ethernet priority.
 *	fl_mlid		- Force loopback and source LID for IB.
 *	rlid		- Remote LID
 *	rmac		- Remote MAC
 *	tclass		- GRH tclass/IPv6 tclass/IPv4 ToS
 *	hop_limit	- GRH hop limit/IPv6 hop limit/IPv4 TTL
 *	grh_gid_fi	- GRH, source GID address and IPv6 flow label.
 *	rgid		- Remote GID/IP address.
 */
static XSCDV_ALWAYS_INLINE
void xscdv_set_dgram_seg(struct xsc_wqe_datagram_seg *seg,
			  uint64_t key, uint32_t dqp_dct,
			  uint8_t ext, uint8_t stat_rate_sl,
			  uint8_t fl_mlid, uint16_t rlid,
			  uint8_t *rmac, uint8_t tclass,
			  uint8_t hop_limit, uint32_t grh_gid_fi,
			  uint8_t *rgid)
{

	/* Always put 64 bits, in q_key, the reserved part will be 0 */
	seg->av.key.dc_key	= htobe64(key);
	seg->av.dqp_dct		= htobe32(((uint32_t)ext << 31) | dqp_dct);
	seg->av.stat_rate_sl	= stat_rate_sl;
	seg->av.fl_mlid		= fl_mlid;
	seg->av.rlid		= htobe16(rlid);
	memcpy(seg->av.rmac, rmac, 6);
	seg->av.tclass		= tclass;
	seg->av.hop_limit	= hop_limit;
	seg->av.grh_gid_fl	= htobe32(grh_gid_fi);
	memcpy(seg->av.rgid, rgid, 16);
}

/*
 * Eth Segment - contains packet headers and information for stateless L2, L3, L4 offloading.
 *
 * Output:
 *	 seg		 - Eth segment to be filled.
 * Input:
 *	cs_flags	 - l3cs/l3cs_inner/l4cs/l4cs_inner.
 *	mss		 - Maximum segment size. For TSO WQEs, the number of bytes
 *			   in the TCP payload to be transmitted in each packet. Must
 *			   be 0 on non TSO WQEs.
 *	inline_hdr_sz	 - Length of the inlined packet headers.
 *	inline_hdr_start - Inlined packet header.
 */
static XSCDV_ALWAYS_INLINE
void xscdv_set_eth_seg(struct xsc_wqe_eth_seg *seg, uint8_t cs_flags,
			uint16_t mss, uint16_t inline_hdr_sz,
			uint8_t *inline_hdr_start)
{
	seg->cs_flags		= cs_flags;
	seg->mss		= htobe16(mss);
	seg->inline_hdr_sz	= htobe16(inline_hdr_sz);
	memcpy(seg->inline_hdr_start, inline_hdr_start, inline_hdr_sz);
}

enum xscdv_set_ctx_attr_type {
	XSCDV_CTX_ATTR_BUF_ALLOCATORS = 1,
};

enum {
	XSC_MMAP_GET_REGULAR_PAGES_CMD	= 0,
	XSC_MMAP_GET_NC_PAGES_CMD	= 3,
};

struct xscdv_ctx_allocators {
	void *(*alloc)(size_t size, void *priv_data);
	void (*free)(void *ptr, void *priv_data);
	void *data;
};

/*
 * Generic context attributes set API
 *
 * Returns 0 on success, or the value of errno on failure
 * (which indicates the failure reason).
 */
int xscdv_set_context_attr(struct ibv_context *context,
		enum xscdv_set_ctx_attr_type type, void *attr);

struct xscdv_clock_info {
	uint64_t nsec;
	uint64_t last_cycles;
	uint64_t frac;
	uint32_t mult;
	uint32_t shift;
	uint64_t mask;
};

/*
 * Get xsc core clock info
 *
 * Output:
 *      clock_info  - clock info to be filled
 * Input:
 *      context     - device context
 *
 * Return: 0 on success, or the value of errno on failure
 */
int xscdv_get_clock_info(struct ibv_context *context,
			  struct xscdv_clock_info *clock_info);

/*
 * Translate device timestamp to nano-sec
 *
 * Input:
 *      clock_info  - clock info to be filled
 *      device_timestamp   - timestamp to translate
 *
 * Return: nano-sec
 */
static inline uint64_t xscdv_ts_to_ns(struct xscdv_clock_info *clock_info,
				       uint64_t device_timestamp)
{
	uint64_t delta, nsec;

	/*
	 * device_timestamp & cycles are the free running 'mask' bit counters
	 * from the hardware hca_core_clock clock.
	 */
	delta = (device_timestamp - clock_info->last_cycles) & clock_info->mask;
	nsec  = clock_info->nsec;

	/*
	 * Guess if the device_timestamp is more recent than
	 * clock_info->last_cycles, if not (too far in the future) treat
	 * it as old time stamp. This will break every max_clock_info_update_nsec.
	 */

	if (delta > clock_info->mask / 2) {
		delta = (clock_info->last_cycles - device_timestamp) &
				clock_info->mask;
		nsec -= ((delta * clock_info->mult) - clock_info->frac) >>
				clock_info->shift;
	} else {
		nsec += ((delta * clock_info->mult) + clock_info->frac) >>
				clock_info->shift;
	}

	return nsec;
}

enum xscdv_context_attr_flags {
	XSCDV_CONTEXT_FLAGS_DEVX = 1 << 0,
};

struct xscdv_context_attr {
	uint32_t flags; /* Use enum xscdv_context_attr_flags */
	uint64_t comp_mask;
};

struct ibv_context *
xscdv_open_device(struct ibv_device *device, struct xscdv_context_attr *attr);

struct xscdv_devx_obj;

struct xscdv_devx_obj *
xscdv_devx_obj_create(struct ibv_context *context, const void *in, size_t inlen,
		       void *out, size_t outlen);
int xscdv_devx_obj_query(struct xscdv_devx_obj *obj, const void *in, size_t inlen,
			  void *out, size_t outlen);
int xscdv_devx_obj_modify(struct xscdv_devx_obj *obj, const void *in, size_t inlen,
			   void *out, size_t outlen);
int xscdv_devx_obj_destroy(struct xscdv_devx_obj *obj);
int xscdv_devx_general_cmd(struct ibv_context *context, const void *in, size_t inlen,
			    void *out, size_t outlen);

struct xscdv_devx_umem {
	uint32_t umem_id;
};

struct xscdv_devx_umem *
xscdv_devx_umem_reg(struct ibv_context *ctx, void *addr, size_t size, uint32_t access);
int xscdv_devx_umem_dereg(struct xscdv_devx_umem *umem);
int xscdv_devx_query_eqn(struct ibv_context *context, uint32_t vector,
			  uint32_t *eqn);

#ifdef __cplusplus
}
#endif

#endif /* _XSCDV_H_ */
